@@ -1,9 +1,12 @@
-# -*- coding: utf-8 -*-
-import unittest
+from collections import namedtuple
+from unittest.mock import patch
 
 from skoolkittest import SkoolKitTestCase
+from skoolkit import components
 from skoolkit.disassembler import Disassembler
 
+Config = namedtuple('Config', 'asm_hex asm_lower defb_size defm_size defw_size')
+
 ASM = {
     '000000': ('NOP', 'NOP', 'NOP'),
     '010000': ('LD BC,0',),
@@ -1800,11 +1803,12 @@ ASM = {
 }
 
 class DisassemblerTest(SkoolKitTestCase):
-    def _get_disassembler(self, snapshot=(), defb_size=8, defb_mod=8, zfill=False, defm_width=66, asm_hex=False, asm_lower=False):
-        return Disassembler(snapshot, defb_size, defb_mod, zfill, defm_width, asm_hex, asm_lower)
+    def _get_disassembler(self, snapshot=(), asm_hex=False, asm_lower=False, defw_size=1):
+        config = Config(asm_hex, asm_lower, 8, 66, defw_size)
+        return Disassembler(snapshot, config)
 
     def _get_snapshot(self, start, data):
-        snapshot = [0] * 65539
+        snapshot = [0] * 65536
         snapshot[start:start + len(data)] = list(data)
         return snapshot
 
@@ -1812,22 +1816,46 @@ class DisassemblerTest(SkoolKitTestCase):
         start = 65536 - len(data)
         snapshot = self._get_snapshot(start, data)
         disassembler = self._get_disassembler(snapshot)
-        return disassembler.disassemble(start)
+        return disassembler.disassemble(start, 65536, 'n')
 
     def defbs_equal(self, *data):
         instructions = self._get_instructions(*data)
         self.assertEqual(len(instructions), 1)
-        self.assertEqual(instructions[0].operation, 'DEFB {0}'.format(','.join([str(b) for b in data])))
+        self.assertEqual(instructions[0][1], 'DEFB {0}'.format(','.join([str(b) for b in data])))
 
     def test_all_instructions(self):
         sna_prefix = [0] * 16384
         for hex_bytes, ops in ASM.items():
             snapshot = sna_prefix + [int(hex_bytes[i:i + 2], 16) for i in range(0, len(hex_bytes), 2)]
             disassembler = self._get_disassembler(snapshot)
-            instructions = disassembler.disassemble(len(sna_prefix), len(snapshot))
-            operations = tuple([inst.operation for inst in instructions])
+            instructions = disassembler.disassemble(len(sna_prefix), len(snapshot), 'n')
+            operations = tuple([inst[1] for inst in instructions])
             self.assertEqual(operations, ops)
 
+    def test_ld1(self):
+        # 65535 LD A,n
+        self.defbs_equal(62)
+
+    def test_ld2(self):
+        # 65535 LD BC,nn
+        self.defbs_equal(1)
+
+    def test_ld3(self):
+        # 65534 LD BC,nn
+        self.defbs_equal(1, 0)
+
+    def test_ld4(self):
+        # 65534 LD (IX+d),A
+        self.defbs_equal(221, 119)
+
+    def test_ld5(self):
+        # 65534 LD (IX+d),n
+        self.defbs_equal(221, 54)
+
+    def test_ld6(self):
+        # 65533 LD (IX+d),n
+        self.defbs_equal(221, 54, 0)
+
     def test_jr1(self):
         # 65535 JR d
         self.defbs_equal(24)
@@ -1891,8 +1919,8 @@ class DisassemblerTest(SkoolKitTestCase):
         # that there will be no suffix after ED
         instructions = self._get_instructions(221, 237)
         self.assertEqual(len(instructions), 2)
-        self.assertEqual(instructions[0].operation, 'DEFB 221')
-        self.assertEqual(instructions[1].operation, 'DEFB 237')
+        self.assertEqual(instructions[0][1], 'DEFB 221')
+        self.assertEqual(instructions[1][1], 'DEFB 237')
 
     def test_dded2(self):
         # 65533 DDED4300
@@ -1901,59 +1929,61 @@ class DisassemblerTest(SkoolKitTestCase):
         # that there will be no room for the 2-byte suffix after ED43
         instructions = self._get_instructions(221, 237, 67, 0)
         self.assertEqual(len(instructions), 2)
-        self.assertEqual(instructions[0].operation, 'DEFB 221')
-        self.assertEqual(instructions[1].operation, 'DEFB 237,67,0')
+        self.assertEqual(instructions[0][1], 'DEFB 221')
+        self.assertEqual(instructions[1][1], 'DEFB 237,67,0')
 
     def test_defb_range(self):
         snapshot = self._get_snapshot(32768, (97, 98, 99, 94, 96, 65, 66, 67, 127))
         disassembler = self._get_disassembler(snapshot)
 
-        instructions = disassembler.defb_range(32768, 32777)
+        sublengths = ((0, 'n'),)
+        instructions = disassembler.defb_range(32768, 32777, sublengths)
         self.assertEqual(len(instructions), 2)
-        operations = [i.operation for i in instructions]
+        operations = [i[1] for i in instructions]
         self.assertEqual(operations, ['DEFB 97,98,99,94,96,65,66,67', 'DEFB 127'])
 
-        sublengths = [(3, 'T'), (2, None), (3, 'T'), (1, None)]
-        instructions = disassembler.defb_range(32768, 32777, sublengths=sublengths)
+        sublengths = [(3, 'c'), (2, 'n'), (3, 'c'), (1, 'n')]
+        instructions = disassembler.defb_range(32768, 32777, sublengths)
         self.assertEqual(len(instructions), 1)
-        self.assertEqual(instructions[0].operation, 'DEFB "abc",94,96,"ABC",127')
+        self.assertEqual(instructions[0][1], 'DEFB "abc",94,96,"ABC",127')
 
-        sublengths = [(3, 'T'), (2, 'B'), (3, 'X'), (1, None)]
-        instructions = disassembler.defb_range(32768, 32777, sublengths=sublengths)
+        sublengths = [(3, 'c'), (2, 'n'), (3, 'c'), (1, 'n')]
+        instructions = disassembler.defb_range(32768, 32777, sublengths)
         self.assertEqual(len(instructions), 1)
-        self.assertEqual(instructions[0].operation, 'DEFB "abc",94,96,"ABC",127')
+        self.assertEqual(instructions[0][1], 'DEFB "abc",94,96,"ABC",127')
 
     def test_defm_range(self):
         snapshot = self._get_snapshot(32768, (97, 98, 99, 94, 96, 65, 66, 67, 127))
         disassembler = self._get_disassembler(snapshot)
 
-        instructions = disassembler.defm_range(32768, 32771)
+        sublengths = ((0, 'c'),)
+        instructions = disassembler.defm_range(32768, 32771, sublengths)
         self.assertEqual(len(instructions), 1)
-        self.assertEqual(instructions[0].operation, 'DEFM "abc"')
+        self.assertEqual(instructions[0][1], 'DEFM "abc"')
 
-        instructions = disassembler.defm_range(32768, 32777)
-        self.assertEqual(len(instructions), 5)
-        operations = [i.operation for i in instructions]
-        self.assertEqual(operations, ['DEFM "abc"', 'DEFB 94', 'DEFB 96', 'DEFM "ABC"', 'DEFB 127'])
-
-        sublengths = [(3, None), (2, 'B'), (3, None), (1, 'B')]
-        instructions = disassembler.defm_range(32768, 32777, sublengths=sublengths)
+        instructions = disassembler.defm_range(32768, 32777, sublengths)
         self.assertEqual(len(instructions), 1)
-        self.assertEqual(instructions[0].operation, 'DEFM "abc",94,96,"ABC",127')
+        self.assertEqual(instructions[0][1], 'DEFM "abc",94,96,"ABC",127')
 
-        sublengths = [(3, 'T'), (2, 'X'), (3, None), (1, 'B')]
-        instructions = disassembler.defm_range(32768, 32777, sublengths=sublengths)
+        sublengths = [(3, 'c'), (2, 'n'), (3, 'c'), (1, 'n')]
+        instructions = disassembler.defm_range(32768, 32777, sublengths)
         self.assertEqual(len(instructions), 1)
-        self.assertEqual(instructions[0].operation, 'DEFM "abc",94,96,"ABC",127')
+        self.assertEqual(instructions[0][1], 'DEFM "abc",94,96,"ABC",127')
 
-    def test_defm_line(self):
-        disassembler = self._get_disassembler()
+        sublengths = [(3, 'c'), (2, 'n'), (3, 'c'), (1, 'h')]
+        instructions = disassembler.defm_range(32768, 32777, sublengths)
+        self.assertEqual(len(instructions), 1)
+        self.assertEqual(instructions[0][1], 'DEFM "abc",94,96,"ABC",$7F')
 
-        instruction = disassembler.defm_line(0, [34, 72, 101, 108, 108, 111, 34])
-        self.assertEqual(instruction.operation, r'DEFM "\"Hello\""')
+    def test_defm_with_escaped_characters(self):
+        snapshot = self._get_snapshot(30000, (34, 72, 101, 108, 108, 111, 34, 67, 58, 92, 84, 69, 77, 80))
+        disassembler = self._get_disassembler(snapshot)
 
-        instruction = disassembler.defm_line(0, [67, 58, 92, 84, 69, 77, 80])
-        self.assertEqual(instruction.operation, r'DEFM "C:\\TEMP"')
+        instruction = disassembler.defm_range(30000, 30007, ((0, 'c'),))[0]
+        self.assertEqual(instruction[1], r'DEFM "\"Hello\""')
+
+        instruction = disassembler.defm_range(30007, 30014, ((0, 'c'),))[0]
+        self.assertEqual(instruction[1], r'DEFM "C:\\TEMP"')
 
     def test_defw_range(self):
         start = 32768
@@ -1961,37 +1991,192 @@ class DisassemblerTest(SkoolKitTestCase):
         snapshot = self._get_snapshot(start, data)
         disassembler = self._get_disassembler(snapshot)
 
-        instructions = disassembler.defw_range(start, start + len(data), False)
+        sublengths = ((0, 'n'),)
+        instructions = disassembler.defw_range(start, start + len(data), sublengths)
         self.assertEqual(len(instructions), 4)
         i = 0
         for address in range(start, start + len(data), 2):
             defw = instructions[i]
             defw_data = data[i * 2:i * 2 + 2]
-            self.assertEqual(defw.address, address)
-            self.assertEqual(defw.operation, 'DEFW {0}'.format(defw_data[0] + 256 * defw_data[1]))
-            self.assertEqual(defw.bytes, defw_data)
+            self.assertEqual(defw[0], address)
+            self.assertEqual(defw[1], 'DEFW {}'.format(defw_data[0] + 256 * defw_data[1]))
+            self.assertEqual(defw[2], defw_data)
             i += 1
 
-        instructions = disassembler.defw_range(start, start + len(data), True)
+        sublengths = ((8, 'n'),)
+        instructions = disassembler.defw_range(start, start + len(data), sublengths)
         self.assertEqual(len(instructions), 1)
         defw = instructions[0]
-        self.assertEqual(defw.address, start)
-        self.assertEqual(defw.operation, 'DEFW 1,257,514,65283')
-        self.assertEqual(defw.bytes, data)
+        self.assertEqual(defw[0], start)
+        self.assertEqual(defw[1], 'DEFW 1,257,514,65283')
+        self.assertEqual(defw[2], data)
 
-    def test_num_str(self):
-        disassembler = self._get_disassembler(asm_hex=False)
-        self.assertEqual(disassembler.num_str(123), '123')
+    def test_defw_range_with_custom_defw_size(self):
+        snapshot = self._get_snapshot(30000, [1, 0, 1, 1, 2, 2])
+        disassembler = self._get_disassembler(snapshot, defw_size=2)
+        exp_instructions = [
+            (30000, 'DEFW 1,257', [1, 0, 1, 1]),
+            (30004, 'DEFW 514', [2, 2])
+        ]
+        instructions = disassembler.defw_range(30000, 30006, ((0, 'n'),))
+        self.assertEqual(exp_instructions, instructions)
 
-        disassembler = self._get_disassembler(asm_hex=True)
-        self.assertEqual(disassembler.num_str(5), '5')
-        self.assertEqual(disassembler.num_str(10), '$0A')
-        self.assertEqual(disassembler.num_str(256), '$0100')
-        self.assertEqual(disassembler.num_str(5, 1), '$05')
-        self.assertEqual(disassembler.num_str(10, 1), '$0A')
-        self.assertEqual(disassembler.num_str(5, 2), '$0005')
-        self.assertEqual(disassembler.num_str(10, 2), '$000A')
-        self.assertEqual(disassembler.num_str(256, 2), '$0100')
+    def test_defw_range_at_64k_boundary(self):
+        snapshot = self._get_snapshot(65535, [1])
+        disassembler = self._get_disassembler(snapshot)
+        instructions = disassembler.defw_range(65535, 65536, ((0, 'n'),))
+        self.assertEqual([(65535, 'DEFB 1', [1])], instructions)
 
-if __name__ == '__main__':
-    unittest.main()
+    def test_defw_range_with_custom_defw_size_at_64k_boundary(self):
+        snapshot = self._get_snapshot(65529, [1, 0, 1, 1, 2, 2, 3])
+        disassembler = self._get_disassembler(snapshot, defw_size=2)
+        exp_instructions = [
+            (65529, 'DEFW 1,257', [1, 0, 1, 1]),
+            (65533, 'DEFW 514', [2, 2]),
+            (65535, 'DEFB 3', [3])
+        ]
+        instructions = disassembler.defw_range(65529, 65536, ((0, 'n'),))
+        self.assertEqual(exp_instructions, instructions)
+
+    def test_lower_case_conversion(self):
+        snapshot = [
+            0,              # 00000 NOP
+            199,            # 00001 RST 0
+            203, 0,         # 00002 RLC B
+            221, 9,         # 00004 ADD IX,BC
+            253, 9,         # 00006 ADD IY,BC
+            237, 64,        # 00008 IN B,(C)
+            221, 203, 0, 6, # 00010 RLC (IX+0)
+            253, 203, 0, 6  # 00014 RLC (IY+0)
+        ]
+        exp_instructions = (
+            (0, 'nop'),
+            (1, 'rst 0'),
+            (2, 'rlc b'),
+            (4, 'add ix,bc'),
+            (6, 'add iy,bc'),
+            (8, 'in b,(c)'),
+            (10, 'rlc (ix+0)'),
+            (14, 'rlc (iy+0)'),
+        )
+        disassembler = self._get_disassembler(snapshot, asm_lower=True)
+        instructions = disassembler.disassemble(0, 18, 'n')
+        self.assertEqual(len(instructions), len(exp_instructions))
+        for instruction, (address, operation) in zip(instructions, exp_instructions):
+            self.assertEqual(instruction[0], address)
+            self.assertEqual(instruction[1], operation)
+
+    def test_lower_case_conversion_with_character_operands(self):
+        snapshot = [
+            62, 65,          # 00000 LD A,"A"
+            221, 70, 66,     # 00002 LD B,(IX+"B")
+            253, 113, 67,    # 00005 LD (IY+"C"),C
+            221, 54, 34, 68, # 00008 LD (IX+"\""),"D"
+        ]
+        exp_instructions = (
+            (0, 'ld a,"A"'),
+            (2, 'ld b,(ix+"B")'),
+            (5, 'ld (iy+"C"),c'),
+            (8, 'ld (ix+"\\""),"D"'),
+        )
+        disassembler = self._get_disassembler(snapshot, asm_lower=True)
+        instructions = disassembler.disassemble(0, 12, 'c')
+        self.assertEqual(len(instructions), len(exp_instructions))
+        for instruction, (address, operation) in zip(instructions, exp_instructions):
+            self.assertEqual(instruction[0], address)
+            self.assertEqual(instruction[1], operation)
+
+    def test_lower_case_conversion_of_defb_statement(self):
+        snapshot = [65, 255]
+        disassembler = self._get_disassembler(snapshot, asm_lower=True)
+        sublengths = ((1, 'c'), (1, 'h'))
+        instructions = disassembler.defb_range(0, 2, sublengths)
+        self.assertEqual(len(instructions), 1)
+        self.assertEqual(instructions[0][1], 'defb "A",$ff')
+
+    def test_lower_case_conversion_of_defm_statement(self):
+        snapshot = [34, 65, 34, 255]
+        disassembler = self._get_disassembler(snapshot, asm_lower=True)
+        sublengths = ((3, 'c'), (1, 'h'))
+        instructions = disassembler.defm_range(0, 4, sublengths)
+        self.assertEqual(len(instructions), 1)
+        self.assertEqual(instructions[0][1], r'defm "\"A\"",$ff')
+
+    def test_lower_case_conversion_of_defm_statement_no_sublengths(self):
+        snapshot = [65, 66]
+        disassembler = self._get_disassembler(snapshot, asm_lower=True)
+        sublengths = ((0, 'c'),)
+        instructions = disassembler.defm_range(0, 2, sublengths)
+        self.assertEqual(len(instructions), 1)
+        self.assertEqual(instructions[0][1], 'defm "AB"')
+
+    def test_lower_case_conversion_of_defs_statement(self):
+        snapshot = [15] * 10
+        disassembler = self._get_disassembler(snapshot, asm_lower=True)
+        sublengths = ((0, 'h'), (0, 'h'))
+        instructions = disassembler.defs_range(0, 10, sublengths)
+        self.assertEqual(len(instructions), 1)
+        self.assertEqual(instructions[0][1], 'defs $0a,$0f')
+
+    def test_lower_case_conversion_of_defw_statement(self):
+        snapshot = [255, 255]
+        disassembler = self._get_disassembler(snapshot, asm_lower=True)
+        sublengths = ((0, 'h'),)
+        instructions = disassembler.defw_range(0, 2, sublengths)
+        self.assertEqual(len(instructions), 1)
+        self.assertEqual(instructions[0][1], 'defw $ffff')
+
+    @patch.object(components, 'SK_CONFIG', None)
+    def test_custom_operand_formatter(self):
+        custom_formatter = """
+            class CustomFormatter:
+                def __init__(self, config):
+                    pass
+                def format_byte(self, value, base):
+                    return '@{:o}'.format(value)
+                def format_word(self, value, base):
+                    return '{:04X}h'.format(value)
+                def is_char(self, value):
+                    return value == ord('A')
+        """
+        self.write_component_config('OperandFormatter', '*.CustomFormatter', custom_formatter)
+        snapshot = [65, 66]
+        disassembler = self._get_disassembler(snapshot)
+
+        instructions = disassembler.defb_range(0, 2, ((0, 'n'),))
+        self.assertEqual([(0, 'DEFB @101,@102', [65, 66])], instructions)
+
+        instructions = disassembler.defw_range(0, 2, ((0, 'n'),))
+        self.assertEqual([(0, 'DEFW 4241h', [65, 66])], instructions)
+
+        instructions = disassembler.defm_range(0, 2, ((0, 'c'),))
+        self.assertEqual([(0, 'DEFM "A",@102', [65, 66])], instructions)
+
+    @patch.object(components, 'SK_CONFIG', None)
+    def test_custom_operand_formatter_api(self):
+        custom_formatter = """
+            class CustomFormatter:
+                def __init__(self, config):
+                    assert config.asm_hex
+                    assert config.asm_lower is False
+                def format_byte(self, value, base):
+                    assert value == 65
+                    assert base == 'c'
+                    return "'A'"
+                def format_word(self, value, base):
+                    assert value == 16961
+                    assert base == 'h'
+                    return '0'
+                def is_char(self, value):
+                    assert value == 65
+                    return True
+        """
+        self.write_component_config('OperandFormatter', '*.CustomFormatter', custom_formatter)
+        snapshot = [65, 66]
+        disassembler = self._get_disassembler(snapshot, asm_hex=True)
+
+        instructions = disassembler.defw_range(0, 2, ((0, 'h'),))
+        self.assertEqual([(0, 'DEFW 0', [65, 66])], instructions)
+
+        instructions = disassembler.defm_range(0, 1, ((0, 'c'),))
+        self.assertEqual([(0, "DEFM 'A'", [65])], instructions)
